#include <linux/module.h> // needed for modules
#include <linux/kernel.h> // needed for KERN_ALERT
#include <linux/init.h>   // needed for macros
#include <linux/fs.h>     // needed for driver management
#include <asm/uaccess.h>  // needed for copy_to/from_user
#include <linux/time.h>	  // needed for time services
#include <linux/timer.h>  // needed for timer services

#include "unDriver.h"

#define TIMERPERIOD 1250

MODULE_AUTHOR("JP Babau");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("premier driver");

static int myMajor = 0;   // allocation dynamique

struct timeval timeZero;	// instant de demarrage du driver
struct timeval currentTime; 	// instant courant

struct timer_list myTimer;	// timer periodique
static int counter;		// compteur de tick timer
spinlock_t lockCounter;		// protection de counter

// structure specifique au peripherique
typedef struct
{
  int ad;		       // adresse peripherique
  struct semaphore semProtect; // protection des donnees peripheriques
  sensorInfo * info;	       // donnees specifiques
} devInfo;

// tableau statique des peripheriques
devInfo theDevices[MAXDEVICES];

static ssize_t myOpen(struct inode *inode, struct file* file)
{
	//unsigned  int minor  = MINOR(inode->i_rdev);
	unsigned  int minor  = iminor(inode);

	// le minor correspond au rang dans le tableau des peripheriques
	if ((minor>=0) && (minor<MAXDEVICES) )
	{

		// association descripteur peripherique/donnees specifiques
		file->private_data = (devInfo*) &theDevices[minor];

		printk(KERN_ALERT "open() OK sur device %i\n",minor);
		return 0;
	}
	else
	{
		printk(KERN_ALERT "open() KO sur device %i\n",minor);
		return -1;		
	}
}

static ssize_t myRead(struct file *file, char * buf, size_t count, loff_t *ppos)
{
	devInfo * p = (devInfo*) file->private_data;
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);

	down_interruptible(&p->semProtect);

	// recuperation des donnees specifiques
	copy_to_user(buf,(sensorInfo*)(p->info), sizeof(sensorInfo));
			
	up(&p->semProtect);

	printk(KERN_ALERT "read() %i(%i) dans le device %i\n",p->info->value,p->info->cpt,minor);

	return (sizeof(sensorInfo));
}

static ssize_t myWrite(struct file *file, const char * buf, size_t count, loff_t *ppos)
{

	devInfo * p = (devInfo*) file->private_data;
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);

	down_interruptible(&p->semProtect);

	// sauvegarde des donnees peripheriques
	copy_from_user((int*)&(p->info->value), buf, sizeof(int)); 
	
	p->info->cpt++;
	do_gettimeofday(&currentTime);
	p->info->time = currentTime.tv_sec-timeZero.tv_sec;

	spin_lock_irq(&lockCounter);
	p->info->pollData = counter;
	spin_unlock_irq(&lockCounter);

	up(&p->semProtect);
	
	printk(KERN_ALERT "write() % i(%i)(%i) dans le device %i\n",p->info->value,p->info->cpt,p->info->time,minor);

	return (sizeof(int));
}

static ssize_t myRelease(struct inode *inode, struct file* file)
{
	printk(KERN_ALERT "close()\n");
	return 0;
}

static int myIoctl(struct inode * inode, struct file * file, unsigned int cmd, unsigned long arg)
{
	devInfo * p = (devInfo*) file->private_data;
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	int ret;

	switch(cmd)
	{
		case RAZ : p->info->cpt=0;break;
		default : ret = -EINVAL;break;
	}

	printk(KERN_ALERT "ioctl() %i dans le device %i\n",cmd,minor);

	return (ret);
}

static struct file_operations fops =
{
	open : myOpen,
	read : myRead,
	write : myWrite,
	ioctl : myIoctl,
	release : myRelease
};


void periodicCounter(unsigned long arg)
{
	spin_lock_irq(&lockCounter);
	counter++;
	spin_unlock_irq(&lockCounter);

//	printk(KERN_DEBUG "timer Activation : %i\n",counter);	
	
	myTimer.expires += (unsigned long) TIMERPERIOD;
	add_timer(&myTimer);
}

static int __init install(void)
{
	int installCR;
	int i;

	printk(KERN_ALERT "hello\n");

	installCR = register_chrdev(myMajor,DRIVERNAME,&fops);

	if (installCR <0)
	{
		printk(KERN_WARNING "probleme d'installation\n");
		return installCR;
	}

	// mise a jour du majeur
	myMajor = installCR;

	// initialisation des informations peripheriques
	for (i=0;i<MAXDEVICES;i++)
	{
		theDevices[i].info=kmalloc(sizeof(sensorInfo),GFP_KERNEL);	
		theDevices[i].info->cpt = 0;
		theDevices[i].info->value = -1;
		init_MUTEX(&theDevices[i].semProtect);
	}
	spin_lock_init(&lockCounter);		
	
	do_gettimeofday(&timeZero); // memorisation de l'instant de lancement du driver

	init_timer(&myTimer);
	myTimer.expires = jiffies + (unsigned long) TIMERPERIOD;
	myTimer.function = periodicCounter;
	counter = 0;
	add_timer(&myTimer);

	printk(KERN_DEBUG "installation OK, MAJOR NUMBER: %i\n",myMajor);
	return 0;
}

static void __exit desinstall(void)
{
	int i;

	printk(KERN_ALERT "see you\n");

	unregister_chrdev(myMajor,DRIVERNAME);

	for (i=0;i<MAXDEVICES;i++)
	{
		kfree(theDevices[i].info);	
	}

	del_timer(&myTimer);

	printk(KERN_DEBUG "desinstallation OK\n");
}


module_param(myMajor,int,0);
MODULE_PARM_DESC(myMajor,"le numero de majeur en parametre du module");

module_init(install);
module_exit(desinstall);

